/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global alert, document, window */

/**
 * View module.
 *
 * @module app.view
 * @requires {@link app}

 * @namespace app.view
 */

// make sure that "app" namespace is created
window.app = window.app || {};
window.app.view = window.app.view || {};

(function view(app) {
    'use strict';

    /**
     * Active class name.
     *
     * @memberof app.view
     * @private
     * @const {string}
     */
    var ACTIVE_CLASS = 'active',

        /**
         * Application element template.
         *
         * @memberof app.view
         * @private
         * @const {string}
         */
        APP_TEMPLATE = '<img src="{imgURI}"/><span>{name}</span>',

        /**
         * Message to be displayed if no MIME type
         * matching files have been found.
         *
         * @memberof app.view
         * @private
         * @const {string}
         */
        NO_FILES_FOUND_MESSAGE = 'No matching files found.',
        
        /**
         * Message to be displayed if no
         * matching applications have been found.
         *
         * @memberof app.view
         * @private
         * @const {string}
         */
        NO_APPLICATIONS_FOUND_MESSAGE = 'No matching applications found.',

        /**
         * Application element.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        appElement = document.body,

        /**
         * Applications list element.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        applicationsList = appElement.querySelector('#application-list'),

        /**
         * Applications list element swipe data.
         *
         * @memberof app.view
         * @private
         * @type {object}
         */
        applicationsListSwipeData = {
            startX: 0,
            currentX: 0,
            offsetX: 0,
            enabled: false
        },

        /**
         * Index of selected application.
         *
         * @memberof app.view
         * @private
         * @type {number}
         * @default
         */
        selectedAppIndex = 0,

        /**
         * Navigation element.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        navigation = appElement.querySelector('nav'),

        /**
         * Command execute application element.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        commandExecute = appElement.querySelector('.command-execute'),

        /**
         * Send button.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        commandSend = appElement.querySelector('.command-send'),

        /**
         * Applications for current operation.
         *
         * @memberof app.view
         * @private
         * @type {ApplicationElement[]}
         */
        applicationsFound = [],

        /**
         * Current operation.
         *
         * @memberof app.view
         * @private
         * @type {string}
         */
        operation = null,

        /**
         * Email input element.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        emailInput = document.querySelector('[type="email"]'),

        /**
         * Message body element.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        messageBody = document.querySelector('.message-box pre'),

        /**
         * MIME selector popup.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        MIMESelectorPopup = document.querySelector('.mime-selector-popup'),

        /**
         * MIME selector.
         *
         * @memberof app.view
         * @private
         * @type {HTMLElement}
         */
        MIMESelector = document.getElementById('mime-selector');


    /**
     * Application element constructor.
     *
     * @constructor
     * @memberof app.view
     * @private
     * @param {ApplicationInformation} application
     */
    function ApplicationElement(application) {
        this.icon = application.iconPath;
        this.appId = application.id;
        this.name = application.name;

        this.application = application;

        this.elem = document.createElement('li');
        this.elem.innerHTML = APP_TEMPLATE
            .replace('{imgURI}', this.icon)
            .replace('{name}', this.name);
    }

    /**
     * Calls application element constructor with provided application data.
     *
     * @memberof app.view
     * @private
     * @param {ApplicationInformation[]} application
     * @returns {ApplicationElement}
     */
    function createAppElement(application) {
        application = new ApplicationElement(application);

        applicationsList.appendChild(application.elem);
        return application;
    }

    /**
     * Clears applications list.
     *
     * @memberof app.view
     * @private
     */
    function clearApplicationList() {
        applicationsList.innerHTML = '';
        applicationsList.style.left = '';
        applicationsListSwipeData.enabled = false;
    }

    /**
     * Updates applications list.
     *
     * @memberof app.view
     * @private
     */
    function updateApplicationList() {
        applicationsFound = app.model.getApplications();

        if (applicationsFound.length) {
        	clearApplicationList();
        	applicationsFound = applicationsFound.map(createAppElement);
            applicationsFound[0].elem.classList.add(ACTIVE_CLASS);

            if (applicationsFound.length > 1) {
                applicationsListSwipeData.enabled = true;
            }
        } else {
            alert(NO_APPLICATIONS_FOUND_MESSAGE);
        }
    }

    /**
     * Shows popup with MIME selector.
     *
     * @memberof app.view
     * @private
     */
    function showMimePopup() {
        MIMESelector.value = 0;
        MIMESelector.focus();
        MIMESelectorPopup.classList.add('active');
    }

    /**
     * Hides popup with MIME selector.
     *
     * @memberof app.view
     * @private
     */
    function hideMimePopup() {
        MIMESelectorPopup.classList.remove('active');
    }

    /**
     * Handles navigation element click event.
     * Calls application model to update current operation.
     *
     * @memberof app.view
     * @private
     * @param {MouseEvent} event
     * @fires "view.operation.changed"
     */
    function onNavClick(event) {
        var target = event.target;

        operation = target.getAttribute('data-operation');

        app.helpers.getElementsInArray('li', navigation)
            .forEach(function removeClass(element) {
                element.classList.remove(ACTIVE_CLASS);
            });
        target.classList.add(ACTIVE_CLASS);

        if (operation === 'view') {
            clearApplicationList();
            showMimePopup();
        } else {
            MIMESelector.value = '';
            triggerOperationChange();
        }
    }

    /**
     * Handles applications list touchstart event.
     * Sets initial swipe coordinates.
     *
     * @memberof app.view
     * @private
     * @param {TouchEvent} e
     */
    function onTouchStartAppList(e) {
        if (!applicationsListSwipeData.enabled) {
            return;
        }

        applicationsListSwipeData.startX =
            e.touches[0].pageX + selectedAppIndex * (window.innerWidth / 2);
        applicationsListSwipeData.currentX = applicationsListSwipeData.startX;

        applicationsList.classList.add('swipe-event');
    }

    /**
     * Handles applications list touchmove event.
     * Sets offset of applications list element.
     *
     * @memberof app.view
     * @private
     * @param {TouchEvent} e
     */
    function onTouchMoveAppList(e) {
        if (!applicationsListSwipeData.enabled) {
            return;
        }

        applicationsListSwipeData.currentX = e.touches[0].pageX;

        applicationsListSwipeData.offsetX =
            applicationsListSwipeData.currentX -
                applicationsListSwipeData.startX;

        applicationsListSwipeData.offsetX = Math.max(
            -(applicationsFound.length) * window.innerWidth / 2,
            Math.min(
                0,
                applicationsListSwipeData.offsetX
            )
        );

        applicationsList.style.left = applicationsListSwipeData.offsetX + 'px';
    }

    /**
     * Handles applications list touchend event.
     * Snaps application list element to closest application element.
     *
     * @memberof app.view
     * @private
     */
    function onTouchEndAppList() {
        if (!applicationsListSwipeData.enabled) {
            return;
        }

        selectedAppIndex = Math.max(
            Math.min(
                applicationsFound.length - 1,
                Math.round(
                    -applicationsListSwipeData.offsetX / (window.innerWidth / 2)
                )
            ),
            0
        );

        applicationsList.classList.remove('swipe-event');

        applicationsList.style.left =
                -selectedAppIndex * window.innerWidth / 2 + 'px';

        applicationsFound.forEach(function removeClass(app) {
            app.elem.classList.remove(ACTIVE_CLASS);
        });

        applicationsFound[selectedAppIndex].elem.classList.add(ACTIVE_CLASS);
    }

    /**
     * Returns selected application index.
     *
     * @memberof app.view
     * @private
     * @returns {number} selectedAppIndex
     */
    function getSelectedApplicationIndex() {
        return selectedAppIndex;
    }

    /**
     * Handles command execute element click.
     *
     * @memberof app.view
     * @private
     */
    function onClickCommandExecute() {
        if (applicationsFound.length) {
            app.model.launchApplication(getSelectedApplicationIndex());    		
        } else {
            alert(NO_APPLICATIONS_FOUND_MESSAGE);
        }
    }

    /**
     * Handles event fired when no images have been found.
     * Shows related message.
     *
     * @memberof app.view
     * @private
     */
    function onNoImagesFound() {
        alert(NO_FILES_FOUND_MESSAGE);
    }

    /**
     * Handles "send" element click.
     *
     * @memberof app.view
     * @private
     */
    function onClickSend() {
        if (app.helpers.validateEmailField(emailInput)) {
            app.model.sendEmail({
                to: emailInput.value,
                body: messageBody.innerText
            });
        } else {
            alert('Email address is required');
        }
    }

    /**
     * Creates MIME selector options.
     *
     * @memberof app.view
     * @private
     */
    function addMIMESelectorOptions() {
        var mimeTypes = app.model.getMimeTypes();

        Object.keys(mimeTypes).forEach(function addOption(mime) {
            var option = document.createElement('option');

            option.value = mime;
            option.innerText = mime;

            MIMESelector.appendChild(option);
        });
    }

    /**
     * Handles MIME selector change event.
     *
     * @memberof app.view
     * @private
     */
    function onChangeMIMESelector() {
        triggerOperationChange();
        hideMimePopup();
    }

    /**
     * Calls model operation change.
     *
     * @memberof app.view
     * @private
     */
    function triggerOperationChange() {
        app.model.changeOperation({
            operation: operation,
            mime: MIMESelector.value
        });
        selectedAppIndex = 0;
    }

    /**
     * Binds events.
     *
     * @memberof app.view
     * @private
     */
    function bindEvents() {
        app.helpers.getElementsInArray('li', navigation)
            .forEach(function addListener(element) {
                element.addEventListener('click', onNavClick, false);
            });

        applicationsList.addEventListener('touchstart', onTouchStartAppList);
        applicationsList.addEventListener('touchmove', onTouchMoveAppList);
        applicationsList.addEventListener('touchend', onTouchEndAppList);

        commandExecute.addEventListener('click', onClickCommandExecute);
        commandSend.addEventListener('click', onClickSend);

        MIMESelector.addEventListener('change', onChangeMIMESelector);

        document.addEventListener(
            'model.applications.list.updated',
            updateApplicationList
        );
        
        document.addEventListener(
            'model.no.matching.applications.found',
            updateApplicationList
        );

        document.addEventListener(
            'model.no.images',
            onNoImagesFound
        );
    }

    /**
     * Returns popup opening state.
     *
     * @memberof app.view
     * @private
     * @returns {boolean}
     */
    function isMimePopupOpen() {
        return MIMESelectorPopup.classList.contains(ACTIVE_CLASS);
    }

    /**
     * Initializes view.
     *
     * @memberof app.view
     * @public
     */
    function init() {
        addMIMESelectorOptions();
        bindEvents();
        clearApplicationList();
    }

    app.view = {
        init: init,
        hideMimePopup: hideMimePopup,
        isMimePopupOpen: isMimePopupOpen
    };

}(window.app));
